package edu.uiuc.dsi.mutex;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.PriorityQueue;

import edu.uiuc.dsi.Packet;
import edu.uiuc.dsi.TCPSocket;
import edu.uiuc.dsi.clocks.LamportClock;

public class Maekawa extends Base
{
	private static final byte MESSAGE_REQUEST = (byte)0x01;
	private static final byte MESSAGE_REPLY = (byte)0x02;
	private static final byte MESSAGE_RELEASE = (byte)0x03;
	private static final byte MESSAGE_FAILED = (byte)0x04;
	private static final byte MESSAGE_INQUIRE = (byte)0x05;
	private static final byte MESSAGE_YIELD = (byte)0x06;

	private LamportClock objLamportClock = new LamportClock();
	private ArrayList<TCPSocket> objTCPSocketsUnreplied = new ArrayList<TCPSocket>();
	private ArrayList<TCPSocket> objRequestSet = new ArrayList<TCPSocket>();
	private LamportQueueElement objCurrentElementWithReply = null;
	private PriorityQueue<LamportQueueElement> objTCPSocketsDeferredReplies = new PriorityQueue<LamportQueueElement>();
	private ArrayList<TCPSocket> objTCPSocketsFailed = new ArrayList<TCPSocket>();
	private ArrayList<TCPSocket> objTCPSocketsHaveInquired = new ArrayList<TCPSocket>();
	
	public Maekawa(String componentName)
	{
		super(componentName);
	}

	@Override
	protected void doBeginMutExTasks()
	{
		objLamportClock.tick();
		
		objTCPSocketsHaveInquired.clear(); // TODO: This check shouldn't be necessary. Donno if it's stable though... do all INQUIREs eventually get FAILED? I don't know.
	}

	@Override
	public void doReleaseMutExTasks()
	{
		objLamportClock.tick();
		
		ByteBuffer objByteBuffer = ByteBuffer.allocate(1);
		objByteBuffer.put(MESSAGE_RELEASE);
		for (TCPSocket objTCPSocket : objRequestSet)
			write(objTCPSocket, objByteBuffer.array());

		objCurrentElementWithReply = objTCPSocketsDeferredReplies.poll();
		if (objCurrentElementWithReply != null)
		{
			ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(3);
			objResponseByteBuffer.put(MESSAGE_REPLY);
			objResponseByteBuffer.putShort(objLamportClock.getCurrentTime());
			write(objCurrentElementWithReply.getSiteTCPSocket(), objResponseByteBuffer.array());						
		}
	}

	@Override
	public void doRequestMutExTasks()
	{
		objLamportClock.tick();
		
		// 1. If no custom request set is defined, define a square-based request set
		ArrayList<TCPSocket> objActiveTCPSocketsList = new ArrayList<TCPSocket>();
		objActiveTCPSocketsList.add(getSelfTCPSocket());
		for (TCPSocket objTCPSocket : getActiveTCPSockets())
			objActiveTCPSocketsList.add(objTCPSocket);
		Collections.sort(objActiveTCPSocketsList);

		int indexOfSelf = Collections.binarySearch(objActiveTCPSocketsList, getSelfTCPSocket());
		int gridSize = (int)Math.ceil(Math.sqrt(objActiveTCPSocketsList.size()));
		
		objRequestSet = new ArrayList<TCPSocket>();
		objRequestSet.add(this.getSelfTCPSocket());
		for (int i=0; i < gridSize; i++)
		{
			int elementToGet;
			
			// Get all elements in the same row...
			elementToGet = ((indexOfSelf / gridSize) * gridSize) + i;
			if (elementToGet < objActiveTCPSocketsList.size() && elementToGet != indexOfSelf)
				objRequestSet.add(objActiveTCPSocketsList.get(elementToGet));
			
			// Get all elements in the same column...
			elementToGet = (indexOfSelf % gridSize) + (gridSize * i);
			if (elementToGet < objActiveTCPSocketsList.size() && elementToGet != indexOfSelf)
				objRequestSet.add(objActiveTCPSocketsList.get(elementToGet));
		}
		
		// 2. Send out REQUESTs to all elements in the RequestSet 
		ByteBuffer objByteBuffer = ByteBuffer.allocate(3);
		objByteBuffer.put(MESSAGE_REQUEST);
		objByteBuffer.putShort(objLamportClock.getCurrentTime());

		for (TCPSocket objTCPSocket : objRequestSet)
			objTCPSocketsUnreplied.add(objTCPSocket);

		for (TCPSocket objTCPSocket : objRequestSet)
			write(objTCPSocket, objByteBuffer.array());
	}

	@Override
	public void processPacket(TCPSocket sender, Packet packet)
	{
		ByteBuffer objByteBuffer = ByteBuffer.wrap(packet.getData());
		byte packetType = objByteBuffer.get();
		short packetTimestamp = 0;
		
		switch (packetType)
		{
			case MESSAGE_REQUEST:
				// Update Clock...
				packetTimestamp = objByteBuffer.getShort();
				objLamportClock.tick(packetTimestamp);
	
				if (objCurrentElementWithReply == null)
				{
					objCurrentElementWithReply = new LamportQueueElement(packetTimestamp, sender);
					
					ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(3);
					objResponseByteBuffer.put(MESSAGE_REPLY);
					objResponseByteBuffer.putShort(objLamportClock.getCurrentTime());
					write(sender, objResponseByteBuffer.array());
				}
				else
				{
					LamportQueueElement objCurrentPacketQueueElement = new LamportQueueElement(packetTimestamp, sender);
					objTCPSocketsDeferredReplies.add(objCurrentPacketQueueElement);
					
					if (objCurrentElementWithReply.compareTo(objCurrentPacketQueueElement) < 0) 
					{
						ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(1);
						objResponseByteBuffer.put(MESSAGE_FAILED);
						write(sender, objResponseByteBuffer.array());
					}
					else
					{
						ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(1);
						objResponseByteBuffer.put(MESSAGE_INQUIRE);
						write(objCurrentElementWithReply.getSiteTCPSocket(), objResponseByteBuffer.array());
					}
				}
				
				break;
				
			case MESSAGE_REPLY:
				// Update Clock...
				packetTimestamp = objByteBuffer.getShort();
				objLamportClock.tick(packetTimestamp);
				
				objTCPSocketsUnreplied.remove(sender);
				objTCPSocketsFailed.remove(sender);
				
				break;
				
			case MESSAGE_RELEASE:
				objTCPSocketsHaveInquired.clear();                                     // v20071127 update
				objCurrentElementWithReply = objTCPSocketsDeferredReplies.poll();
				
				if (objCurrentElementWithReply != null)
				{
					ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(3);
					objResponseByteBuffer.put(MESSAGE_REPLY);
					objResponseByteBuffer.putShort(objLamportClock.getCurrentTime());
					write(objCurrentElementWithReply.getSiteTCPSocket(), objResponseByteBuffer.array());						
				}
				
				break;
				
			case MESSAGE_FAILED:
				objTCPSocketsFailed.add(sender);
				
				for (TCPSocket objTCPSocket : objTCPSocketsHaveInquired)
				{
					objTCPSocketsFailed.add(objTCPSocket);
					objTCPSocketsUnreplied.add(objTCPSocket);

					ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(1);
					objResponseByteBuffer.put(MESSAGE_YIELD);
					write(objTCPSocket, objResponseByteBuffer.array());	
				}
				
				objTCPSocketsHaveInquired.clear();
				
				break;
				
			case MESSAGE_INQUIRE:
				if (objTCPSocketsFailed.size() > 0)
				{
					objTCPSocketsFailed.add(sender);
					objTCPSocketsUnreplied.add(sender);
					
					ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(1);
					objResponseByteBuffer.put(MESSAGE_YIELD);
					write(sender, objResponseByteBuffer.array());						
				}
				else if (!objTCPSocketsHaveInquired.contains(sender))
					objTCPSocketsHaveInquired.add(sender);
				
				break;
				
			case MESSAGE_YIELD:
				objTCPSocketsDeferredReplies.add(objCurrentElementWithReply);
				objCurrentElementWithReply = objTCPSocketsDeferredReplies.poll();
				
				ByteBuffer objResponseByteBuffer = ByteBuffer.allocate(3);
				objResponseByteBuffer.put(MESSAGE_REPLY);
				objResponseByteBuffer.putShort(objLamportClock.getCurrentTime());
				write(objCurrentElementWithReply.getSiteTCPSocket(), objResponseByteBuffer.array());						
				
				break;
		}
		
		// Execute MutEx if conditions are met...
		if (objTCPSocketsUnreplied.size() == 0 &&
			objCurrentElementWithReply != null &&
			objCurrentElementWithReply.getSiteTCPSocket() == getSelfTCPSocket() &&
			this.getCurrentMutualExclusionState() == Base.REQUESTING_CRITICAL_SECTION)
		{
			try { beginMutEx(); }
			catch (MutExException e) { e.printStackTrace(); }
		}

		
		this.getApplicationCore().writeStatusMessage(this, "Status Update:");

		this.getApplicationCore().writeStatusMessage(this, "  Current State     = " + this.getCurrentMutualExclusionState());
		
		this.getApplicationCore().writeStatusMessage(this, "  Packet Type       = " + packetType);
		this.getApplicationCore().writeStatusMessage(this, "  Lamport Clock     = " + objLamportClock.getCurrentTime());
		this.getApplicationCore().writeStatusMessage(this, "  Packet Sender+Clk = " + sender.getSocketName() + " @ " + packetTimestamp);

		if (objCurrentElementWithReply == null)
			this.getApplicationCore().writeStatusMessage(this, "  Current Element   = null");
		else
			this.getApplicationCore().writeStatusMessage(this, "  Current Element   = " + objCurrentElementWithReply.getSiteTCPSocket().getSocketName() + " @ " + objCurrentElementWithReply.getTimestamp());
		
		
		if (objTCPSocketsDeferredReplies.isEmpty())
			this.getApplicationCore().writeStatusMessage(this, "  Next Element      = null");
		else
			this.getApplicationCore().writeStatusMessage(this, "  Next Element      = " + objTCPSocketsDeferredReplies.peek().getSiteTCPSocket().getSocketName() + " @ " + objTCPSocketsDeferredReplies.peek().getTimestamp());
		
		this.getApplicationCore().writeStatusMessage(this, "  Sockets Unreplied (" + objTCPSocketsUnreplied.size() + ")");
		for (int i=0; i < objTCPSocketsUnreplied.size(); i++)
			this.getApplicationCore().writeStatusMessage(this, "    " + objTCPSocketsUnreplied.get(i).getSocketName());

		this.getApplicationCore().writeStatusMessage(this, "  Sockets Failed (" + objTCPSocketsFailed.size() + ")");
		for (int i=0; i < objTCPSocketsFailed.size(); i++)
			this.getApplicationCore().writeStatusMessage(this, "    " + objTCPSocketsFailed.get(i).getSocketName());
		
		this.getApplicationCore().writeStatusMessage(this, "  Request Set (" + objRequestSet.size() + ")");
		for (int i=0; i < objRequestSet.size(); i++)
			this.getApplicationCore().writeStatusMessage(this, "    " + objRequestSet.get(i).getSocketName());

		this.getApplicationCore().writeStatusMessage(this, "  Have Inquired (" + objTCPSocketsHaveInquired.size() + ")");
		for (int i=0; i < objTCPSocketsHaveInquired.size(); i++)
			this.getApplicationCore().writeStatusMessage(this, "    " + objTCPSocketsHaveInquired.get(i).getSocketName());

		
		
		
	}
}